from peliqan.client.base import BaseClient


class SFTPClient(BaseClient):
    def __init__(self, connection, jwt, backend_url):
        super(SFTPClient, self).__init__(jwt, backend_url)
        self.connection = connection

    def sftp_via_proxy(self, path, action, stream=False, **kwargs):
        payload = {
            "connection": self.connection,
            "path": path,
            "action": action,
            "kwargs": kwargs
        }
        url = f"{self.BACKEND_URL}/api/proxy/sftp/"
        return self.call_backend('post', url, stream=stream, json=payload)

    def _write_to_file(self, chunk_iter, encoding, file_object):
        for chunk in chunk_iter:
            chunked_data = chunk
            if encoding:
                chunked_data = chunk.decode(encoding)
            file_object.write(chunked_data)
        return file_object

    def _write_to_memory(self, chunk_iter, encoding):
        chunked_data_array = []
        for chunk in chunk_iter:
            chunked_data = chunk
            if encoding:
                chunked_data = chunk.decode(encoding)

            chunked_data_array.append(chunked_data)

        return {
            'status': 'success',
            'detail': "".join(chunked_data_array) if encoding else b"".join(chunked_data_array)
        }

    def read_file(self, path, encoding='utf-8', file_object=None, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        chunk_iter = self.sftp_via_proxy(path, 'read_file', stream=True, **kwargs)

        if file_object:
            return self._write_to_file(chunk_iter, encoding, file_object)
        else:
            return self._write_to_memory(chunk_iter, encoding)

    def dir(self, path, *args, **kwargs):
        kwargs = self.args_to_kwargs(args, kwargs)
        response = self.sftp_via_proxy(path, 'dir', **kwargs)
        return response



class PeliqanClientException(Exception):
    """
    Base exception raised by the Peliqan module.
    """

    def __init__(self, message, code=None, error_dict=None, *args, **kwargs):
        self.message = message
        self.code = code
        self.error_dict = error_dict

    @property
    def error(self):
        return {
            'code': self.code,
            'message': self.message,
            'error_dict': self.error_dict,
        }


class OperationNotSupported(PeliqanClientException):
    """
        Raise this when an operation is not support by the client.
    """


class PeliqanJsonSerializerException(PeliqanClientException):
    """
        Raise this when a json encoding fails for a data structure.
    """



import ast
import json

import requests

from peliqan.exceptions import PeliqanClientException
from peliqan.utils import _serialize_data


class BaseClient:

    def __init__(self, jwt, backend_url):
        self.JWT = jwt
        self.BACKEND_URL = backend_url

    def get_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": "JWT %s" % self.JWT
        }

    def _get_stream_iter(self, response):
        try:
            for chunk in response.iter_content(chunk_size=32768):
                yield chunk
        except Exception as e:
            raise PeliqanClientException(f"Error while streaming data. Original error is {e}")

    def _parse_error_response_dict(self, response_dict):
        if not isinstance(response_dict, dict):
            return None, None, response_dict

        error_code = response_dict.get("error")
        details = response_dict.get('detail', '')

        try:
            if not isinstance(details, dict):
                details = ast.literal_eval(details)

                if isinstance(details, tuple):
                    error = details[0]
                    trace_id = details[1]
                    details = {'trace_id': trace_id.strip('TraceId: '), **error}
        except Exception:
            return error_code, None, details

        parsed_response = []
        error_type = details.get('type')
        if error_type:
            parsed_response.append(error_type)

        operation = details.get('operation')
        if operation:
            parsed_response.append(f"during {operation}:")

        message = details.get('message')
        if message:
            parsed_response.append(message)

        message = details.get('description')
        if message:
            parsed_response.append(message)

        stack_trace = details.get('stack_trace')
        if stack_trace:
            line_details = stack_trace[-1]
            parsed_response.append(f"\nin {line_details}")

        if not parsed_response:
            str_data = json.dumps(details)
            parsed_response.append(str_data)

        return error_code, details, ' '.join(parsed_response)

    def call_backend(self, method, url, expected_status_code=200, **kwargs):
        if not kwargs.get('headers'):
            headers = self.get_headers()
            kwargs.update(headers=headers)

        json_data = kwargs.get('json')
        if json_data:
            serialized_data = _serialize_data(json_data)
            kwargs['json'] = serialized_data

        stream = kwargs.get('stream', False)
        response = requests.request(method, url, **kwargs)

        if stream:
            return self._get_stream_iter(response)
        else:
            try:
                response_dict = response.json()
            except ValueError:
                response_dict = {}

        # handle error responses
        if response.status_code != expected_status_code:
            error_message = f"Server responded with status code {response.status_code}. "
            error_code = None
            if not response_dict:
                response_dict = response.text

            error_dict = None
            if response_dict:
                error_code, error_dict, parsed_error_response = self._parse_error_response_dict(response_dict)
                error_message += f"{parsed_error_response}"

            raise PeliqanClientException(error_message, error_code, error_dict)

        return response_dict

    def args_to_kwargs(self, args, kwargs):
        """
        Used to allow using both a dict argument or keyword arguments:
        pq.add("contact", name='John', city='NY') or
        pq.add("contact", contact_obj)
        """
        for arg in args:
            if type(arg) != dict:
                raise PeliqanClientException("Only arguments of type dict and kwargs are accepted")
            kwargs.update(**arg)
        return kwargs



from peliqan.utils import canonicalize_identifier


def test_canonicalize_identifier():
    test_cases = [
        ("asda\"'.", "asda___"),
        ("As_mf", "as_mf"),
        ("Ömer Hayyam", "ömer_hayyam"),
        ("Pøstmark", "pøstmark"),
    ]
    for test_case in test_cases:
        result = canonicalize_identifier(test_case[0])
        assert result == test_case[1]



from openai import OpenAI
from peliqan.client import BaseClient

from peliqan.exceptions import PeliqanClientException


class AIClient(BaseClient):
    def __init__(self, connection, jwt, backend_url):
        super().__init__(jwt, backend_url)
        self.connection = connection

    def _get_open_ai_client(self, data):
        api_key = data.get('password')
        return OpenAI(api_key=api_key)

    def get_ai_connection_data(self):
        url = f"{self.BACKEND_URL}/connection/ai/config/?connection={self.connection}"
        return self.call_backend('get', url)

    def get_client(self):
        data = self.get_ai_connection_data()
        provider = data["connector"]
        if provider == 'openai':
            return self._get_open_ai_client(data)
        else:
            PeliqanClientException("Provider {} is not currently supported.".format(provider))
